-----------------------------------------------------------------------------

local MAJOR_VERSION = "TagCompiler-1.0"
local MINOR_VERSION = tonumber(("$Revision: 38489 $"):match("%d+")) or 0

-- This ensures the code is only executed if the libary doesn't already exist, or is a newer version
if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary.") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

-----------------------------------------------------------------------------
--
-- TextTag			:= { Statement } @
-- RawTag			:= RValue @
-- NumberTag		:= RValue @
--
-- Statement		:= literal | "[" RValue "]"
-- RValue			:= Logical
-- Logical			:= Relation { "&" | "|" } Relation
-- Relation			:= Expression [ ("==" | "~=" | ">" | "<" | "<=" | ">=") Expression ]
-- Expression		:= Term { ("+" | "-") Term }
-- Factor			:= Chain { ("*" | "/") Chain }
-- Chain			:= Unary { ":" Unary }
-- Unary			:= [ "~" | "-" ] Primary
-- Primary			:= ConstantNumber | ConstantString | Identifier ["(" ParameterList ")"] | "(" RValue ")"
-- ParameterList	:= RValue {, RValue }
--
-----------------------------------------------------------------------------

local TagCompiler = { }

-----------------------------------------------------------------------------
--
-- TagCompiler:Compile(tag, tagData, resultType)
-- tagData is a structure that optionally has:
--  tagname.data
--  tagname.events
--  tagname.overrideEvents
--  tagname.guaranteeResult
--  tagname.noGuaranteeChange
--  tagname.numeric
--  parameterTypes				-- Array of either "raw", "number" or "string"
-- And must have one of:
--  tagname.tableData
--  tagname.string
--  tagname.method				-- will be called with the input parameter, prior result and any parameters
--  tagname.compileString
--
-- Note that tagname must all be lower case
--
-- resultType is "string", "number", or "raw"
--
-- the result of CreateTag is a table with
--  data		- a list of data elements required, which is all of the tag's .data lists
--  events		- a list of the events required, which is a list of all tag's .events minus .overrideEvents
--  method		- call with input parameter
--  tag			- the tag used
--
-----------------------------------------------------------------------------

TagCompiler.CommonTagData = { }

TagCompiler.CommonTagData.angle = 
{
	method = function(data, value) return value and value ~= "" and ("<%s>"):format(value) end,
	noGuaranteeChange = true
}

TagCompiler.CommonTagData.bracket =
{
	method = function(data, value) return value and value ~= "" and ("(%s)"):format(value) end,
	noGuaranteeChange = true
}

TagCompiler.CommonTagData.brace = 
{
	method = function(data, value)	return value and value ~= "" and ("{%s}"):format(value) end,
	noGuaranteeChange = true
}

TagCompiler.CommonTagData.color =
{
	method = function(data, value, r, g, b) return value and ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value) end,
	noGuaranteeChange = true
}

TagCompiler.CommonTagData.colorif = 
{
	method = function(data, value, condition, r, g, b)
				return value and condition and ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value) or value
			 end,
	noGuaranteeChange = true
}

TagCompiler.CommonTagData.format =
{
	method = function(data, value, format) return value and format:format(value) end,
	noGuaranteeChange = true,
	parameterTypes = { "string" }
}

TagCompiler.CommonTagData.hexcolor =
{
	method = function(data, value, colorValue)	return value and ("|cff%s%s|r"):format(colorValue, value or "") end,
	noGuaranteeChange = true,
	parameterTypes = { "string" }
}

TagCompiler.CommonTagData.hexcolorif = 
{
	method = function(data, value, condition, colorValue) return value and condition and ("|cff%s%s|r"):format(colorValue, value) or value end,
	noGuaranteeChange = true,
	parameterTypes = { "raw", "string" }
}

TagCompiler.CommonTagData.hideif =
{
	compileString = function(dataTable, prior, condition)
						return string.format("not (%s)%s%s", condition, prior and " and " or "", prior or "")
					end,
	parameterTypes = { "raw" }
}

TagCompiler.CommonTagData.hidezero = 
{
	method = function(data, value) return value ~= 0 and value end
}

TagCompiler.CommonTagData.lower =
{
	method = function(data, value)
				if type(value) == "string" then
					return value:lower()
				end
				return value
			 end
}

TagCompiler.CommonTagData.maximum =
{
	method = function(data, value, v1, v2) if v1 > v2 then return v1 else return v2 end end,
	parameterTypes = { "number", "number" }
}

TagCompiler.CommonTagData.minimum =
{
	method = function(data, value, v1, v2) if v1 < v2 then return v1 else return v2 end end,
	parameterTypes = { "number", "number" }
}

TagCompiler.CommonTagData.negate =
{
	method = function(data, value)
				if type(value) == "number" then
					return -value
				end
				return value
			 end,
	noGuaranteeChange = true
}

TagCompiler.CommonTagData["not"] =
{
	compileString = function(dataTable, prior, condition)
						return string.format("not (%s)%s%s", condition, prior and " and " or "", prior or "")
					end,
	parameterTypes = { "raw" }
}

TagCompiler.CommonTagData.percent =
{
	method = function(data, value)
				if type(value) == "number" then
					return value.."%"
				end
				return value
			 end,
	noGuaranteeChange = true
}

TagCompiler.CommonTagData.playername =
{
	compileString = function(dataTable, prior)
						return string.format("%s%s\"%s\"", prior or "", prior and " and " or "", UnitName("player"))
					end,
	guaranteeResult = true,
}

TagCompiler.CommonTagData.prefix =
{
	method = function(data, value, prefix)	return value and prefix..value end,
	noGuaranteeChange = true,
	parameterTypes = { "string" }
}

TagCompiler.CommonTagData.select =
{
	compileString = function(dataTable, prior, condition, trueResult, falseResult)
						return string.format("%s%s((%s) and (%s) or (%s))", prior or "", prior and " and " or "", condition, trueResult, falseResult)
					end,
	parameterTypes = { "raw", "raw", "raw" }
}

TagCompiler.CommonTagData.surround =
{
	method = function(data, value, prefix, suffix)	return value and prefix..value..suffix end,
	noGuaranteeChange = true,
	parameterTypes = { "string", "string" }
}

TagCompiler.CommonTagData.rgb =
{
	method = function(data, value, r, g, b) return value and ("|cff%02x%02x%02x%s|r"):format(r, g, b, value) end,
	parameterTypes = { "number", "number", "number" }
}

TagCompiler.CommonTagData.rgbif =
{
	method = function(data, value, condition, r, g, b)	return value and condition and ("|cff%02x%02x%02x%s|r"):format(r, g, b, value) or value end,
	parameterTypes = { "raw", "number", "number", "number" }
}

TagCompiler.CommonTagData.short =
{
	method = function(data, value)
				if type(value) == "number" then
					-- Use a 'binary' tree to determine our range, rather than the old sequential search
					-- This gives us maximum 3 comparisons
					if value <= -10000 then
						if value <= -100000 then
							if value <= -1000000 then
								return ("%.2fm"):format(value / 1000000)
							else
								return ("%dk"):format(value / 1000)
							end
						else
							return ("%.1fk"):format(value / 1000)
						end
					else
						if value < 100000 then
							if value < 10000 then
								return ("%d"):format(value)
							else
								return ("%.1fk"):format(value / 1000)
							end
						else
							if value < 1000000 then
								return ("%dk"):format(value / 1000)
							else
								return ("%.2fm"):format(value / 1000000)
							end				
						end
					end
				end
				return value
			 end,
	noGuaranteeChange = true
}

TagCompiler.CommonTagData.suffix =
{
	method = function(data, value, suffix)	return value and value..suffix end,
	noGuaranteeChange = true,
	parameterTypes = { "string" }
}

TagCompiler.CommonTagData.upper =
{
	method = function(data, value)
				if type(value) == "string" then
					return value:upper()
				end
				return value
			 end,
	noGuaranteeChange = true
}

TagCompiler.CommonTagData.white		= { string = [[hexColor("ffffff")]] }
TagCompiler.CommonTagData.red 		= { string = [[hexColor("ff0000")]] }
TagCompiler.CommonTagData.green 	= { string = [[hexColor("00ff00")]] }
TagCompiler.CommonTagData.blue 		= { string = [[hexColor("0000ff")]] }
TagCompiler.CommonTagData.cyan 		= { string = [[hexColor("00ffff")]] }
TagCompiler.CommonTagData.fuchsia 	= { string = [[hexColor("ff00ff")]] }
TagCompiler.CommonTagData.yellow 	= { string = [[hexColor("ffff00")]] }
TagCompiler.CommonTagData.gray 		= { string = [[hexColor("afafaf")]] }

-----------------------------------------------------------------------------

local tagCompilerData
local parseTag
local parseTagOffset
local headerCode = { }
local dataTable
local overrideEvents = { }
local numeric = false

local function GetByte()
	while parseTag:byte(parseTagOffset) == 32 do 
		parseTagOffset = parseTagOffset + 1
	end 
	return parseTag:byte(parseTagOffset) or 0
end

local function AdvanceByte()
	parseTagOffset = parseTagOffset + 1
end

-----------------------------------------------------------------------------

local TagErrorMethods =
{
	string	= function() return "!!Tag Error!!" end,
	number	= function() return -1 end,
	raw		= function() return nil end,
}

local function TagErrorNoOffset(x) dataTable.error = true tagCompilerData.TagError(x) end
local function TagError(x) TagErrorNoOffset(("%s near \"%s...\" (%d)"):format(x, parseTag:sub(parseTagOffset, parseTagOffset+3), parseTagOffset)) end
local function TagDebug(x) tagCompilerData.TagDebug(x) end

local function EmptyTable(table) for k in pairs(table) do table[k] = nil end end
local function MergeData(targetTable, sourceData)
	if not sourceData then return end
	if type(sourceData) == "string" then
		targetTable[sourceData] = true
	elseif type(sourceData) == "table" then
		for _,v in pairs(sourceData) do
			targetTable[v] = true
		end
	end
end

-----------------------------------------------------------------------------

local ParseRValue

local function CoerceToNumber(text)
	if not text or text == "" then
		numeric = true
		guaranteeResult = true
		return "0"
	end
	
	if guaranteeResult then
		if numeric then return text end
		numeric = true
		return ("(tonumber(%s) or 0)"):format(text)
	else
		guaranteeResult = true
		if numeric then
			return ("(%s or 0)"):format(text)
		else
			numeric = true
			return ("(tonumber(%s) or 0)"):format(text)
		end
	end
end

local function CoerceToString(text)
	if guaranteeResult then
		return text
	else
		guaranteeResult = true
		if not text then return [[""]] end
		return ([[(%s or "")]]):format(text)
	end
end

local function CoerceToRaw(text)
	return text
end

local coerceTypeMethods =
{
	string	= CoerceToString,
	number	= CoerceToNumber,
	raw		= CoerceToRaw,
}

local function CoerceToType(text, type, identifier)
	if not type then
		TagErrorNoOffset("Too many parameters for "..identifier)
		return ""
	end

	local method = coerceTypeMethods[type]
	if not method then
		TagError("Internal error - Unrecognised parameter type: "..type)
		return ""
	end

	return method(text)
end

local emptyParameterList = { }
local function ParseParameterList(parameterTypes, identifier)
	if GetByte() == 40 then -- byte == '('
		AdvanceByte()
		local byte = GetByte()
		if byte == 41 then 
			AdvanceByte()
			if parameterTypes then
				TagErrorNoOffset("Missing parameters for: "..identifier)
			end
			return emptyParameterList
		end

		if not parameterTypes then
			TagErrorNoOffset("No parameters required for: "..identifier)
			return emptyParameterList
		end

		local parameterListStringList = { }
		table.insert(parameterListStringList, CoerceToType(ParseRValue(), parameterTypes[1], identifier))
		while parseTagOffset <= #parseTag and GetByte() == 44 do -- byte == ','
			AdvanceByte()
			table.insert(parameterListStringList, CoerceToType(ParseRValue(), parameterTypes[#parameterListStringList+1], identifier))
		end

		if GetByte() == 41 then -- byte == ')'
			AdvanceByte()
		else
			TagError("Missing end of parameter list \")\"")
		end
		
		if #parameterListStringList ~= #parameterTypes then
			TagErrorNoOffset(("%s requires %d parameters, only given %d"):format(identifier, #parameterTypes, #parameterListStringList))
		end

		return parameterListStringList
	else
		if parameterTypes then
			TagErrorNoOffset("Missing parameters for: "..identifier)
		end
	end
	return emptyParameterList
end

local function ParsePrimary(functionString)
	local byte = GetByte()
	if byte >= 48 and byte <= 59 then -- byte >= '0' && byte <= '9'
		local startOffset = parseTagOffset
		local endOffset = startOffset
		local decimalPointFound = false
		while true do
			endOffset = endOffset + 1
			byte = parseTag:byte(endOffset) or 0
			if byte == 46 then -- '.'
				if decimalPointFound then
					break
				else
					decimalPointFound = true
				end
			elseif byte < 48 or byte > 59 then
				break
			end
		end
		parseTagOffset = endOffset
		if functionString then
			numeric = false
			return ("(%s and %s)"):format(functionString, parseTag:sub(startOffset, endOffset-1))
		else
			numeric = true
			guaranteeResult = true
			return parseTag:sub(startOffset, endOffset-1)
		end
	elseif byte == 34 then	-- byte == '\"'
		local startOffset = parseTagOffset+1
		local endOffset = startOffset + 1
		numeric = false
		while true do
			if endOffset > #parseTag then
				TagError("Missing closing quote")
				return ""
			end
			byte = parseTag:byte(endOffset) or 0
			if byte == 34 then -- byte == '\"'
				break
			elseif byte == 92 then	-- byte == '\\'
				endOffset = endOffset + 2
			else
				endOffset = endOffset + 1
			end
		end
		parseTagOffset = endOffset+1
		local string = parseTag:sub(startOffset, endOffset-1):gsub("\\\"", "\"")
		if functionString then
			return ("%s and %q"):format(functionString, string)
		else
			guaranteeResult = true
			return ("%q"):format(string)
		end		
	elseif (byte >= 65 and byte <= 90) or (byte >= 97 and byte <= 122) then	-- byte >= 'A' && byte <= 'Z' || byte >= 'a' && byte <= 'z'
		local startOffset = parseTagOffset
		local endOffset = startOffset
		while (byte >= 65 and byte <= 90) or (byte >= 97 and byte <= 122) do
			endOffset = endOffset + 1
			byte = parseTag:byte(endOffset) or 0
		end
		parseTagOffset = endOffset
		local identifier = parseTag:sub(startOffset, endOffset-1):lower()

		local tagData = tagCompilerData[identifier]
		if not tagData then
			TagErrorNoOffset("Invalid tag: "..identifier)
			return ""
		end

		MergeData(dataTable.data, tagData.data)
		MergeData(dataTable.events, tagData.events)
		MergeData(overrideEvents, tagData.overrideEvents)

		if tagData.string then
			local tag, offset = parseTag, parseTagOffset
			parseTag, parseTagOffset = tagData.string, 1
			local returnValue = ParsePrimary(functionString)
			parseTag, parseTagOffset = tag, offset
			return returnValue
		end

		local result
		local oldGuaranteeResult = guaranteeResult

		if tagData.tableData then
			if functionString then
				result = functionString.." and data."..tagData.tableData
			else
				result = "data."..tagData.tableData
			end
		elseif tagData.method then
			headerCode[("local tagMethod%s = TagCompiler.tagCompilerData[\"%s\"].method\n"):format(identifier, identifier)] = true
			local parameterStringList = ParseParameterList(tagData.parameterTypes, identifier)
			if not dataTable.error then
				if #parameterStringList > 0 then
					result = ("tagMethod%s(data, %s, %s)"):format(identifier, functionString or "true", table.concat(parameterStringList, ", "))
				else
					result = ("tagMethod%s(data, %s)"):format(identifier, functionString or "true")
				end
			end
		elseif tagData.compileString then
			local parameterStringList = ParseParameterList(tagData.parameterTypes, identifier)
			if not dataTable.error then
				result = tagData.compileString(dataTable, functionString, unpack(parameterStringList))
			end
		end

		numeric = tagData.numeric
		if tagData.noGuaranteeChange then
			guaranteeResult = oldGuaranteeResult
		else
			guaranteeResult = tagData.guaranteeResult
		end
		
		return result		
	elseif (byte == 40) then -- '('
		AdvanceByte()
		local priorNumeric = numeric
		local priorGuaranteeResult = guaranteeResult
		local result = ParseRValue()

		local byte = GetByte()
		if byte == 41 then
			AdvanceByte()
			if functionString then
				numeric = numeric and priorNumeric
				guaranteeResult = guaranteeResult and priorGuaranteeResult
				return ("%s and (%s)"):format(functionString, result)
			else
				return ("(%s)"):format(result)
			end
		else
			TagError("Missing \")\"")
			return ""
		end
	else
		TagError("Syntax error")
		return ""
	end
end

local function ParseUnary(functionString)
	local byte = GetByte()
	if byte == 126 then -- '~'
		AdvanceByte()
		local result = ("not (%s)%s%s"):format(ParsePrimary(), functionString and " and " or "", functionString or "")
		guaranteeResult = false
		numeric = false
		return result
	elseif byte == 45 then -- '-'
		AdvanceByte()
		return ("%s%s-%s"):format(functionString or "", functionString and " and " or "", CoerceToNumber(ParsePrimary()))
	else
		return ParsePrimary(functionString)
	end
end

local function ParseChain()
	local functionString = ParseUnary()
	while GetByte() == 58 do	-- while tag[offset] == ':'
		AdvanceByte()
		functionString = ParseUnary(functionString)
	end
	return functionString
end

local function ParseFactor()
	local result = ParseChain()
	local byte = GetByte()
	while byte == 42 or byte == 47 do	-- '*' or '/'
		AdvanceByte()
		result = CoerceToNumber(result)
		result = ("%s %s %s"):format(result, string.char(byte), CoerceToNumber(ParseFactor()))
		numeric = true
		guaranteeResult = true
		byte = GetByte()
	end
	return result
end

local function ParseExpression()
	local result = ParseFactor()
	local byte = GetByte()
	while byte == 43 or byte == 45 do	-- '+' or '-'
		AdvanceByte()
		result = CoerceToNumber(result)
		result = ("(%s %s %s)"):format(result, string.char(byte), CoerceToNumber(ParseFactor()))
		numeric = true
		guaranteeResult = true
		byte = GetByte()
	end
	return result
end

local function ParseRelation()
	local result = ParseExpression()

	local byte = GetByte()
	local byte2 = parseTag:byte(parseTagOffset+1)
	if byte == 61 then	-- '='
		AdvanceByte()
		if byte2 == 61 then
			AdvanceByte()
			result = ("(%s == %s)"):format(result, ParseExpression())
			guaranteeResult = false
			numeric = false
			return result
		else
			TagError("Unexpected \"=\"")
			return ""
		end
	elseif byte == 126 then -- '~'
		AdvanceByte()
		if byte2 == 61 then
			AdvanceByte()
			result = ("(%s ~= %s)"):format(result, ParseExpression())
			guaranteeResult = false
			numeric = false
			return result
		else
			TagError("Unexpected \"=\"")
			return ""
		end
	elseif byte == 60 then -- '<'
		AdvanceByte()
		result = CoerceToNumber(result)
		if byte2 == 61 then
			AdvanceByte()
			result = ("(%s <= %s)"):format(result, CoerceToNumber(ParseExpression()))
			guaranteeResult = false
			numeric = false
			return result
		else
			result = ("(%s < %s)"):format(result, CoerceToNumber(ParseExpression()))
			guaranteeResult = false
			numeric = false
			return result
		end
	elseif byte == 62 then -- '>'
		AdvanceByte()
		result = CoerceToNumber(result)
		if byte2 == 61 then
			AdvanceByte()
			result = ("(%s >= %s)"):format(result, CoerceToNumber(ParseExpression()))
			guaranteeResult = false
			numeric = false
			return result
		else
			result = ("(%s > %s)"):format(result, CoerceToNumber(ParseExpression()))
			guaranteeResult = false
			numeric = false
			return result
		end
	else
		return result
	end
end

local function ParseLogical()
	local result = ParseRelation()
	
	local byte = GetByte()
	while byte == 38 or byte == 124 do	-- '&' or '|'
		AdvanceByte()
		local priorNumeric = numeric
		local priorGuaranteeResult = guaranteeResult
		
		if byte == 38 then -- '&'
			result = ("%s and %s"):format(result, ParseRelation())
			numeric = numeric and priorNumeric
			guaranteeResult = guaranteeResult and priorGuaranteeResult
		else
			result = ("%s or %s"):format(result, ParseRelation())
		end
		byte = GetByte()
	end
	
	return result
end

ParseRValue = function()
	guaranteeResult = false
	return ParseLogical()
end

-----------------------------------------------------------------------------

local headerTable = { }
local functionStringList = { }
function TagCompiler:Compile(tag, tagData, resultType)
	parseTag = tag or ""
	tagCompilerData = tagData

	parseTagOffset = 1
	dataTable = { data = { }, events = { }, tag = parseTag, headerCode = headerCode }

	table.insert(functionStringList, "") -- it's replaced with all the locals required later
	table.insert(functionStringList, "return function(data)\nreturn ")

	if resultType == "string" then
		while parseTagOffset <= #parseTag and not dataTable.error do
			if parseTagOffset > 1 then
				table.insert(functionStringList, "..\n ")
			end
			if parseTag:byte(parseTagOffset) == 91 then	-- if tag[offset] == '['
				AdvanceByte()
				table.insert(functionStringList, CoerceToType(ParseRValue(), "string"))
				if GetByte() == 93 then -- if tag[offset] == ']'
					AdvanceByte()
				elseif not dataTable.error then
					TagError("Tag not closed - missing \"]\"")
				end
			else
				-- literal section
				local startOffset = parseTagOffset
				local endOffset = startOffset+1
				while endOffset <= #parseTag and parseTag:byte(endOffset) ~= 91 do
					endOffset = endOffset + 1
				end
				table.insert(functionStringList, ("%q"):format(parseTag:sub(startOffset, endOffset-1)))
				parseTagOffset = endOffset
			end
		end
	else
		table.insert(functionStringList, CoerceToType(ParseRValue(), resultType))
		if GetByte() ~= 0 then
			TagError("Tag contains unused trailing characters")
		end
	end

	for k in pairs(overrideEvents) do
		dataTable.events[k] = nil
		overrideEvents[k] = nil
	end

	if dataTable.error then
		dataTable.method = TagErrorMethods[resultType]
		EmptyTable(dataTable.data)
		EmptyTable(dataTable.events)
	else
		self.tagCompilerData = tagData
		headerTable[1] = "local TagCompiler = AceLibrary(\""..MAJOR_VERSION.."\")\n"
		for k in pairs(headerCode) do
			table.insert(headerTable, k)
		end
		functionStringList[1] = table.concat(headerTable)
		EmptyTable(headerTable)
		table.insert(functionStringList, "\nend")
		local functionString = table.concat(functionStringList)

		if tagData.TagDebug then
			TagDebug(string.format("Tag: %s", parseTag))
			for event in pairs(dataTable.events) do
				TagDebug(string.format(" - event: %s", event))
			end
			for data in pairs(dataTable.data) do
				TagDebug(string.format(" - data: %s", data))
			end
			TagDebug("--- Compiled method ---\n"..functionString)
		end

		local method = loadstring(functionString)
		dataTable.method = method and method() or TagErrorMethods[resultType]
	end

	EmptyTable(headerCode)
	EmptyTable(functionStringList)

	dataTable.headerCode = nil

	return dataTable
end

-----------------------------------------------------------------------------

AceLibrary:Register(TagCompiler, MAJOR_VERSION, MINOR_VERSION)

-----------------------------------------------------------------------------
